use std::sync::Arc;

use crate::{
    racert::{CertificateBuilder, RaTlsCertificate, RaTlsCertificateBuilder},
    RaTlsConfig, RaTlsConfigBuilder, SgxError,
};
use rustls::client::danger::HandshakeSignatureValid;
use rustls::crypto::{aws_lc_rs, verify_tls12_signature, verify_tls13_signature, CryptoProvider};
use rustls::pki_types::{CertificateDer, ServerName, UnixTime};
use rustls::{
    client::{danger::ServerCertVerified, danger::ServerCertVerifier, ResolvesClientCert},
    sign::CertifiedKey,
    ClientConfig, DigitallySignedStruct, Error, SignatureScheme,
};

#[derive(Debug)]
pub struct RaTlsServerCertVerifier {
    config: RaTlsConfig,
    provider: CryptoProvider,
}

impl RaTlsServerCertVerifier {
    pub fn new(config: RaTlsConfig) -> Self {
        Self {
            config,
            provider: aws_lc_rs::default_provider(), // FIPS certified provider
        }
    }
}

impl ServerCertVerifier for RaTlsServerCertVerifier {
    fn verify_server_cert(
        &self,
        end_entity: &CertificateDer,
        _intermediates: &[CertificateDer],
        _server_name: &ServerName,
        _ocsp_response: &[u8],
        _now: UnixTime,
    ) -> Result<ServerCertVerified, Error> {
        end_entity
            .verify_quote(&self.config)
            .map_err(|e| Error::General(e.to_string()))?;

        Ok(ServerCertVerified::assertion())
    }

    // TLS 1.2 signature verification from the AWS libcrypto
    fn verify_tls12_signature(
        &self,
        message: &[u8],
        cert: &CertificateDer<'_>,
        dss: &DigitallySignedStruct,
    ) -> Result<HandshakeSignatureValid, Error> {
        let supported_schemes = self.provider.signature_verification_algorithms;
        verify_tls12_signature(message, cert, dss, &supported_schemes)
    }

    // TLS 1.3 signature verification from the AWS libcrypto
    fn verify_tls13_signature(
        &self,
        message: &[u8],
        cert: &CertificateDer<'_>,
        dss: &DigitallySignedStruct,
    ) -> Result<HandshakeSignatureValid, Error> {
        let supported_schemes = self.provider.signature_verification_algorithms;
        verify_tls13_signature(message, cert, dss, &supported_schemes)
    }

    fn supported_verify_schemes(&self) -> Vec<SignatureScheme> {
        self.provider
            .signature_verification_algorithms
            .supported_schemes()
    }
}

#[derive(Debug)]
pub struct RaTlsClientCertResolver {
    cert: Arc<CertifiedKey>,
}

impl RaTlsClientCertResolver {
    pub fn new() -> Result<Self, SgxError> {
        let builder = RaTlsCertificateBuilder::new().with_common_name("Client".to_string());
        let cert = builder.build().map(Arc::new)?;
        Ok(Self { cert })
    }
}

impl ResolvesClientCert for RaTlsClientCertResolver {
    fn resolve(
        &self,
        _acceptable_issuers: &[&[u8]],
        _sigschemes: &[rustls::SignatureScheme],
    ) -> Option<Arc<CertifiedKey>> {
        Some(self.cert.clone())
    }

    fn has_certs(&self) -> bool {
        true
    }
}

impl RaTlsConfigBuilder<ClientConfig> for ClientConfig {
    fn from_ratls_config(config: RaTlsConfig) -> Result<Self, SgxError> {
        Ok(Self::builder()
            .dangerous()
            .with_custom_certificate_verifier(Arc::new(RaTlsServerCertVerifier::new(config)))
            .with_client_cert_resolver(Arc::new(RaTlsClientCertResolver::new()?)))
    }
}
