use crate::bindings::*;
use crate::error::SgxError;
use crate::quote::{ReportData, VerifyResult};
use log::{trace, warn};
use std::time::Instant;

type HandleType = *mut ::std::os::raw::c_void;

// TODO: it should be refactored. makes it abstract and hide quote_size and supplemental_size
pub struct SgxIoctlClient {
    fd: HandleType,
    quote_size: Option<u32>,
    supplemental_size: Option<u32>,
}

// Needed to numb the compiler
unsafe impl Send for SgxIoctlClient {}

impl SgxIoctlClient {
    pub(crate) fn new() -> Self {
        Self {
            fd: std::ptr::null_mut(),
            quote_size: None,
            supplemental_size: None,
        }
    }

    pub(crate) fn handle(&mut self) -> Result<HandleType, SgxError> {
        if self.fd.is_null() {
            let handle = unsafe { dcap_quote_open() };
            if handle.is_null() {
                return Err(SgxError::DcapError(
                    "Failed to open DCAP quote device".to_string(),
                ));
            }
            self.fd = handle;
        }
        Ok(self.fd)
    }

    fn get_quote_size(&mut self) -> Result<u32, SgxError> {
        if self.quote_size.is_none() {
            let size = unsafe { dcap_get_quote_size(self.handle()?) };
            trace!("DCAP quote size is {}", size);
            self.quote_size = Some(size);
        }

        Ok(*self.quote_size.as_ref().unwrap())
    }

    fn get_supplemental_data_size(&mut self) -> Result<u32, SgxError> {
        if self.supplemental_size.is_none() {
            let size = unsafe { dcap_get_supplemental_data_size(self.handle()?) };
            trace!("DCAP supplemental data size is {}", size);
            self.supplemental_size = Some(size);
        }

        Ok(*self.supplemental_size.as_ref().unwrap())
    }

    pub fn generate_quote(&mut self, report_data: ReportData) -> Result<Vec<u8>, SgxError> {
        let instant = Instant::now();
        let quote_buf = self.generate_quote_inner(report_data)?;
        trace!("Generated quote in {:?}ms", instant.elapsed().as_millis());
        Ok(quote_buf)
    }

    fn generate_quote_inner(&mut self, report_data: ReportData) -> Result<Vec<u8>, SgxError> {
        let quote_size = self.get_quote_size()?;
        let mut quote_buf: Vec<u8> = vec![0; quote_size as usize];

        let instant = Instant::now();
        let ret_code = unsafe {
            dcap_generate_quote(self.handle()?, quote_buf.as_mut_ptr(), &report_data.into())
        };
        trace!("Generated quote in {:?}ms", instant.elapsed().as_millis());

        if ret_code < 0 {
            return Err(SgxError::DcapError(
                "Failed to generate DCAP quote".to_string(),
            ));
        }

        Ok(quote_buf)
    }

    pub fn verify_quote(&mut self, quote_buf: &[u8]) -> Result<VerifyResult, SgxError> {
        let instant = Instant::now();
        let result = self.verify_quote_inner(quote_buf)?;
        trace!("Verified quote in {:?}ms", instant.elapsed().as_millis());

        if result.is_negligible() {
            if !result.is_ok() {
                warn!("DCAP quote verification returned: {:?}", result);
            }
            return Ok(result);
        }

        Err(SgxError::QuoteError(format!(
            "DCAP quote verification returned: {:?}",
            result
        )))
    }

    fn verify_quote_inner(&mut self, quote_buf: &[u8]) -> Result<VerifyResult, SgxError> {
        let supplemental_data_size = self.get_supplemental_data_size()?;

        let mut status = 1;
        let mut result = SGX_QL_QV_RESULT_UNSPECIFIED;
        let mut suppl_buf: Vec<u8> = vec![0; supplemental_data_size as usize];

        let ret_code = unsafe {
            dcap_verify_quote(
                self.handle()?,
                quote_buf.as_ptr(),
                quote_buf.len() as u32,
                &mut status,
                &mut result,
                supplemental_data_size,
                suppl_buf.as_mut_ptr(),
            )
        };

        if ret_code < 0 {
            return Err(SgxError::DcapError(
                "Failed to verify DCAP quote".to_string(),
            ));
        }

        Ok(result.into())
    }
}

impl Drop for SgxIoctlClient {
    fn drop(&mut self) {
        unsafe {
            if !self.fd.is_null() {
                dcap_quote_close(self.fd);
            }
        }
    }
}
