use crate::bindings::*;
use crate::error::SgxError;
use crate::ioctl::SgxIoctlClient;
#[cfg(feature = "occlum")]
use crate::quote::{Quote, IOCTL_CLIENT, STATIC_QUOTE};
use crate::sgx_consts::{
    SGX_FLAGS_KSS, SGX_KEYPOLICY_CONFIGID, SGX_KEYPOLICY_ISVEXTPRODID, SGX_KEYPOLICY_ISVFAMILYID,
    SGX_KEYPOLICY_MRENCLAVE, SGX_KEYPOLICY_MRSIGNER, SGX_KEYSELECT_SEAL,
    SGX_KEY_REQUEST_RESERVED2_BYTES, TSEAL_DEFAULT_FLAGSMASK, TSEAL_DEFAULT_MISCMASK,
};

#[allow(dead_code)]
pub enum SealingKeyPolicy {
    MrSigner,
    MrEnclave,
}

pub type Sgx128BitKey = sgx_key_128bit_t;

#[derive(Debug)]
pub struct SgxSealing {
    #[cfg(feature = "occlum")]
    pub sealing_key: Sgx128BitKey,
}

impl SgxSealing {
    #[cfg(feature = "occlum")]
    pub(crate) fn get_current_sgx_quote() -> Result<&'static Quote, SgxError> {
        Ok(STATIC_QUOTE.as_ref().map_err(|e| e.clone())?)
    }

    #[cfg(feature = "occlum")]
    pub fn new() -> Result<Self, SgxError> {
        // by default, use MrEnclave key policy
        Self::with_policy(SealingKeyPolicy::MrEnclave)
    }

    pub fn with_policy(policy: SealingKeyPolicy) -> Result<Self, SgxError> {
        let quote = Self::get_current_sgx_quote()?;
        let sealing_key = IOCTL_CLIENT
            .lock()
            .unwrap()
            .generate_sealing_key(quote, policy)?;
        Ok(Self { sealing_key })
    }

    #[cfg(not(feature = "occlum"))]
    pub fn new() -> Result<Self, SgxError> {
        Ok(Self {})
    }
}

impl SgxIoctlClient {
    /// Generate a sealing key for the given policy and SGX report
    /// The sealing key is used to encrypt/decrypt data in the enclave
    /// The quote must be previously generated using the `generate_quote`
    fn generate_sealing_key(
        &mut self,
        quote: &Quote,
        policy: SealingKeyPolicy,
    ) -> Result<Sgx128BitKey, SgxError> {
        let report_body = unsafe { *quote.report_body };
        let mut key_policy = match policy {
            SealingKeyPolicy::MrSigner => SGX_KEYPOLICY_MRSIGNER,
            SealingKeyPolicy::MrEnclave => SGX_KEYPOLICY_MRENCLAVE,
        };

        if (report_body.attributes.flags & SGX_FLAGS_KSS) != 0 {
            const KEY_POLICY_KSS: uint16_t =
                SGX_KEYPOLICY_CONFIGID | SGX_KEYPOLICY_ISVFAMILYID | SGX_KEYPOLICY_ISVEXTPRODID;
            key_policy = key_policy | KEY_POLICY_KSS;
        }

        // Intel sgx sdk 1.8
        let attribute_mask = sgx_attributes_t {
            flags: TSEAL_DEFAULT_FLAGSMASK,
            xfrm: 0,
        };

        let key_id = sgx_key_id_t::default();

        let misc_mask: sgx_misc_select_t = TSEAL_DEFAULT_MISCMASK;

        let mut key = sgx_key_128bit_t::default();
        let key_request = sgx_key_request_t {
            key_name: SGX_KEYSELECT_SEAL,
            key_policy,
            isv_svn: report_body.isv_svn,
            reserved1: 0_u16,
            cpu_svn: report_body.cpu_svn,
            attribute_mask,
            key_id,
            misc_mask,
            config_svn: report_body.config_svn,
            reserved2: [0_u8; SGX_KEY_REQUEST_RESERVED2_BYTES],
        };

        let ret_code = unsafe { utils_gen_key(self.handle()?, &mut key, &key_request) };
        if ret_code < 0 {
            return Err(SgxError::DcapError(
                "Failed to generate DCAP sealing key".to_string(),
            ));
        }

        Ok(key)
    }
}
